@@ -3,11 +3,14 @@ import datetime
 import hashlib
 import importlib
 import inspect
+import functools
 import json
 import logging
 import os, stat, errno
 import pickle
 import shutil
+from typing import Dict, List, Union, Callable
+from types import ModuleType
 
 import networkx as nx
 import yaml
@@ -40,6 +43,8 @@ class StageInstance:
         self.instance = instance
         self.name = name
         self.module_hash = module_hash
+        if not hasattr(self.instance, "execute"):
+            raise RuntimeError("Stage %s does not have execute method" % self.name)
 
     def parameterize(self, parameters):
         return ParameterizedStage(self.instance, self.name)
@@ -55,49 +60,53 @@ class StageInstance:
         return None
 
     def execute(self, context):
-        if hasattr(self.instance, "execute"):
-            return self.instance.execute(context)
-        else:
-            raise RuntimeError("Stage %s does not have execute method" % self.name)
+        return self.instance.execute(context)
 
-def get_module_hash(module):
-    if os.path.exists(module.__file__):
-        with open(module.__file__) as f:
-            hash = hashlib.md5()
-            hash.update(f.read().encode("utf-8"))
-            return hash.hexdigest()
 
-    return None
+def get_stage_hash(descriptor):
+    source = inspect.getsource(descriptor)
+    hash = hashlib.md5()
+    hash.update(source.encode("utf-8"))
+    return hash.hexdigest()
+
 
 def resolve_stage(descriptor):
-    module_hash = None
-
+    # Supported descriptors: module, class, @stage-decorated function or stage-looking object
     if isinstance(descriptor, str):
+        # If a string, first try to get the actual object
         try:
-            # Try to get the module referenced by the string
             descriptor = importlib.import_module(descriptor)
-            module_hash = get_module_hash(descriptor)
         except ModuleNotFoundError:
-            # Not a module, but maybe a class?
-            parts = descriptor.split(".")
-
-            module = importlib.import_module(".".join(parts[:-1]))
-            module_hash = get_module_hash(module)
-
-            constructor = getattr(module, parts[-1])
-            descriptor = constructor()
+            try:
+                parts = descriptor.split(".")
+                module = importlib.import_module(".".join(parts[:-1]))
+                descriptor = getattr(module, parts[-1])
+            except (ValueError, ModuleNotFoundError):
+                return None  # definitely not a stage
 
     if inspect.ismodule(descriptor):
-        module_hash = get_module_hash(descriptor)
-        return StageInstance(descriptor, descriptor.__name__, module_hash)
+        stage_hash = get_stage_hash(descriptor)
+        return StageInstance(descriptor, descriptor.__name__, stage_hash)
 
     if inspect.isclass(descriptor):
-        module = importlib.import_module(descriptor.__module__)
-        module_hash = get_module_hash(module)
-        return StageInstance(descriptor(), "%s.%s" % (descriptor.__module__, descriptor.__name__), module_hash)
+        stage_hash = get_stage_hash(descriptor)
+        return StageInstance(descriptor(), "%s.%s" % (descriptor.__module__, descriptor.__name__), stage_hash)
 
-    clazz = descriptor.__class__
-    return StageInstance(descriptor, "%s.%s" % (clazz.__module__, clazz.__name__), module_hash)
+    if inspect.isfunction(descriptor):
+        if not hasattr(descriptor, 'stage_params'):
+            raise PipelineError("Functions need to be decorated with @synpp.stage in order to be used in the pipeline.")
+        function_stage = DecoratedStage(execute_func=descriptor, stage_params=descriptor.stage_params)
+        stage_hash = get_stage_hash(descriptor)
+        return StageInstance(function_stage, "%s.%s" % (descriptor.__module__, descriptor.__name__), stage_hash)
+
+    if hasattr(descriptor, 'execute'):
+        # Last option: arbitrary object which looks like a stage
+        clazz = descriptor.__class__
+        stage_hash = get_stage_hash(clazz)
+        return StageInstance(descriptor, "%s.%s" % (clazz.__module__, clazz.__name__), stage_hash)
+
+    # couldn't resolve stage (this is something else)
+    return None
 
 def get_config_path(name, config):
     if name in config:
@@ -159,7 +168,8 @@ class ConfiguredStage:
         self.hashed_name = hash_name(instance.name, configuration_context.required_config)
 
     def configure(self, context):
-        return self.instance.configure(context)
+        if hasattr(self.instance, "configure"):
+            return self.instance.configure(context)
 
     def execute(self, context):
         return self.instance.execute(context)
@@ -168,9 +178,14 @@ class ConfiguredStage:
         return self.instance.validate(context)
 
 
-class ConfigurationContext:
-    def __init__(self, base_config):
+class Context:
+    pass
+
+
+class ConfigurationContext(Context):
+    def __init__(self, base_config, config_definitions=[]):
         self.base_config = base_config
+        self.config_requested_stages = [resolve_stage(d).instance for d in config_definitions]
 
         self.required_config = {}
 
@@ -205,8 +220,14 @@ class ConfigurationContext:
             if not alias is None:
                 self.aliases[alias] = definition
 
+    def stage_is_config_requested(self, descriptor):
+        if self.config_requested_stages:
+            return resolve_stage(descriptor).instance in self.config_requested_stages
+        else:
+            return False
 
-class ValidateContext:
+
+class ValidateContext(Context):
     def __init__(self, required_config, cache_path):
         self.required_config = required_config
         self.cache_path = cache_path
@@ -221,7 +242,7 @@ class ValidateContext:
         return self.cache_path
 
 
-class ExecuteContext:
+class ExecuteContext(Context):
     def __init__(self, required_config, required_stages, aliases, working_directory, dependencies, cache_path, pipeline_config, logger, cache, dependency_info):
         self.required_config = required_config
         self.working_directory = working_directory
@@ -324,6 +345,8 @@ def process_stages(definitions, global_config):
 
         # Resolve the underlying code of the stage
         wrapper = resolve_stage(definition["descriptor"])
+        if wrapper is None:
+            raise PipelineError(f"{definition['descriptor']} is not a supported object for pipeline stage definition!")
 
         # Call the configure method of the stage and obtain parameters
         config = copy.copy(global_config)
@@ -332,7 +355,7 @@ def process_stages(definitions, global_config):
             config.update(definition["config"])
 
         # Obtain configuration information through configuration context
-        context = ConfigurationContext(config)
+        context = ConfigurationContext(config, [d['descriptor'] for d in definitions])
         wrapper.configure(context)
 
         definition = copy.copy(definition)
@@ -481,15 +504,19 @@ def update_json(meta, working_directory):
     shutil.move("%s/pipeline.json.new" % working_directory, "%s/pipeline.json" % working_directory)
 
 
-def run(definitions, config = {}, working_directory = None, flowchart_path = None, dryrun = False, verbose = False, logger = logging.getLogger("synpp")):
+def run(definitions, config = {}, working_directory = None, flowchart_path = None, dryrun = False, verbose = False,
+        logger = logging.getLogger("synpp"), rerun_required=True, ensure_working_directory=False):
     # 0) Construct pipeline config
     pipeline_config = {}
     if "processes" in config: pipeline_config["processes"] = config["processes"]
     if "progress_interval" in config: pipeline_config["progress_interval"] = config["progress_interval"]
 
-    if not working_directory is None:
-        if not os.path.isdir(working_directory):
-            raise PipelineError("Working directory does not exist: %s" % working_directory)
+    if ensure_working_directory and working_directory is None:
+        working_directory = '.synpp_cache'
+    if working_directory is not None:
+        if not os.path.isdir(working_directory) and ensure_working_directory:
+            logger.warning("Working directory does not exist, it will be created: %s" % working_directory)
+            os.mkdir(working_directory)
 
         working_directory = os.path.realpath(working_directory)
 
@@ -580,9 +607,11 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
 
     # 4) Devalidate stages
     sorted_cached_hashes = sorted_hashes - ephemeral_counts.keys()
+    stale_hashes = set()
 
-    # 4.1) Devalidate if they are required
-    stale_hashes = set(required_hashes)
+    # 4.1) Devalidate if they are required (optional, otherwise will reload from cache)
+    if rerun_required:
+        stale_hashes.update(required_hashes)
 
     # 4.2) Devalidate if not in meta
     for hash in sorted_cached_hashes:
@@ -592,7 +621,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
     # 4.3) Devalidate if configuration values have changed
     # This devalidation step is obsolete since we have implicit config parameters
 
-    # 4.3) Devalidate if module hash of a stage has changed
+    # 4.4) Devalidate if module hash of a stage has changed
     for hash in sorted_cached_hashes:
         if hash in meta:
             if not "module_hash" in meta[hash]:
@@ -605,7 +634,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
                 if previous_module_hash != current_module_hash:
                     stale_hashes.add(hash)
 
-    # 4.3) Devalidate if cache is not existant
+    # 4.5) Devalidate if cache is not existant
     if not working_directory is None:
         for hash in sorted_cached_hashes:
             directory_path = "%s/%s.cache" % (working_directory, hash)
@@ -614,7 +643,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
             if not hash in cache_available:
                 stale_hashes.add(hash)
 
-    # 4.4) Devalidate if parent has been updated
+    # 4.6) Devalidate if parent has been updated
     for hash in sorted_cached_hashes:
         if not hash in stale_hashes and hash in meta:
             for dependency_hash, dependency_update in meta[hash]["dependencies"].items():
@@ -624,7 +653,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
                     if meta[dependency_hash]["updated"] > dependency_update:
                         stale_hashes.add(hash)
 
-    # 4.5) Devalidate if parents are not the same anymore
+    # 4.7) Devalidate if parents are not the same anymore
     for hash in sorted_cached_hashes:
         if not hash in stale_hashes and hash in meta:
             cached_hashes = set(meta[hash]["dependencies"].keys())
@@ -633,7 +662,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
             if not cached_hashes == current_hashes:
                 stale_hashes.add(hash)
 
-    # 4.6) Manually devalidate stages
+    # 4.8) Manually devalidate stages
     for hash in sorted_cached_hashes:
         stage = registry[hash]
         cache_path = "%s/%s.cache" % (working_directory, hash)
@@ -645,13 +674,13 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
         if not validation_token == existing_token:
             stale_hashes.add(hash)
 
-    # 4.7) Devalidate descendants of devalidated stages
+    # 4.9) Devalidate descendants of devalidated stages
     for hash in set(stale_hashes):
         for descendant_hash in nx.descendants(graph, hash):
             if not descendant_hash in stale_hashes:
                 stale_hashes.add(descendant_hash)
 
-    # 4.8) Devalidate ephemeral stages if necessary
+    # 4.10) Devalidate ephemeral stages if necessary
     pending = set(stale_hashes)
 
     while len(pending) > 0:
@@ -761,6 +790,14 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
                 progress, len(stale_hashes), 100 * progress / len(stale_hashes)
             ))
 
+    if not rerun_required:
+        # Load remaining previously cached results
+        for hash in required_hashes:
+            if results[required_hashes.index(hash)] is None:
+                with open("%s/%s.p" % (working_directory, hash), "rb") as f:
+                    logger.info("Loading cache for %s ..." % hash)
+                    results[required_hashes.index(hash)] = pickle.load(f)
+
     if verbose:
         info = {}
 
@@ -778,26 +815,144 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
 
 
 def run_from_yaml(path):
-    with open(path) as f:
-        settings = yaml.load(f, Loader = yaml.SafeLoader)
+    Synpp.build_from_yml(path).run_pipeline()
 
-    definitions = []
 
-    for item in settings["run"]:
-        parameters = {}
+# Convenience class mostly for running stages individually (possibly interactively, e.g. in Jupyter)
+class Synpp:
+    def __init__(self, config: dict, working_directory: str = None, logger: logging.Logger = logging.getLogger("synpp"),
+                 definitions: List[Dict[str, Union[str, Callable, ModuleType]]] = None, flowchart_path: str = None,
+                 dryrun: bool = False):
+        self.config = config
+        self.working_directory = working_directory
+        self.logger = logger
+        self.definitions = definitions
+        self.flowchart_path = flowchart_path
+        self.dryrun = dryrun
 
-        if type(item) == dict:
-            key = list(item.keys())[0]
-            parameters = item[key]
-            item = key
+    def run_pipeline(self, definitions=None, rerun_required=False, dryrun=None, verbose=False, flowchart_path=None):
+        if definitions is None and self.definitions is None:
+            raise PipelineError("A list of stage definitions must be available in object or provided explicitly.")
+        elif definitions is None:
+            definitions = self.definitions
+        if dryrun is None:
+            dryrun = self.dryrun
+        return run(definitions, self.config, self.working_directory, flowchart_path=flowchart_path,
+                   dryrun=dryrun, verbose=verbose, logger=self.logger, rerun_required=rerun_required,
+                   ensure_working_directory=True)
 
-        definitions.append({
-            "descriptor": item, "parameters": parameters
-        })
+    def run_single(self, descriptor, config={}, rerun_if_cached=False, dryrun=False, verbose=False):
+        return run([{'descriptor': descriptor, 'config': config}], self.config, self.working_directory,
+                   dryrun=dryrun, verbose=verbose, logger=self.logger, rerun_required=rerun_if_cached,
+                   flowchart_path=self.flowchart_path, ensure_working_directory=True)[0]
 
-    config = settings["config"] if "config" in settings else {}
-    working_directory = settings["working_directory"] if "working_directory" in settings else None
-    flowchart_path = settings["flowchart_path"] if "flowchart_path" in settings else None
-    dryrun = settings["dryrun"] if "dryrun" in settings else False
+    @staticmethod
+    def build_from_yml(config_path):
+        with open(config_path) as f:
+            settings = yaml.load(f, Loader=yaml.SafeLoader)
 
-    run(definitions=definitions, config=config, working_directory=working_directory, flowchart_path=flowchart_path, dryrun=dryrun)
+        definitions = []
+
+        for item in settings["run"]:
+            parameters = {}
+
+            if type(item) == dict:
+                key = list(item.keys())[0]
+                parameters = item[key]
+                item = key
+
+            definitions.append({
+                "descriptor": item, "config": parameters
+            })
+
+        config = settings["config"] if "config" in settings else {}
+        working_directory = settings["working_directory"] if "working_directory" in settings else None
+        flowchart_path = settings["flowchart_path"] if "flowchart_path" in settings else None
+        dryrun = settings["dryrun"] if "dryrun" in settings else False
+
+        return Synpp(config=config, working_directory=working_directory, definitions=definitions,
+                     flowchart_path=flowchart_path, dryrun=dryrun)
+
+
+def stage(function=None, *args, **kwargs):
+    def decorator(_func):
+        functools.wraps(_func)
+        _func.stage_params = kwargs
+        return _func
+
+    # parameterized decorator
+    if function is None:
+        return decorator
+    # parameterized decorator where a non-function stage is passed
+    # this should be used like @stage(arg=stage("path.stage"))
+    elif not inspect.isfunction(function):
+        stage = resolve_stage(function)
+        if stage is not None:
+            stage.instance.stage_params = kwargs
+            return stage.instance
+        else:
+            raise PipelineError(f'{function} could not be resolved as a stage.')
+    else:  # unparameterized decorator
+        return decorator(function)
+
+
+def get_context():
+    stack = list(frame for frame in inspect.stack())
+    for frame in stack:
+        f_locals = dict(frame.frame.f_locals)
+        for name, obj in f_locals.items():
+            if name == 'context' and isinstance(obj, Context):
+                return obj
+    return None
+
+
+class DecoratedStage:
+    def __init__(self, execute_func: Callable, stage_params: dict):
+        self.execute_func = execute_func
+        self.stage_params = stage_params
+        self.func_params = inspect.signature(self.execute_func).parameters
+
+    def configure(self, context):
+        self._resolve_params(context)
+
+    def execute(self, context):
+        return self.execute_func(**self._resolve_params(context))
+
+    def _resolve_params(self, context):
+        """
+        This function is used both when configuring a stage as well as collecting the kwargs for execution.
+        Stages may be decorated in three different ways:
+        1. simple decoration:
+                @synpp.stage
+                def foo(bar='baz'): ...
+            Which will look for 'bar' in the global context config (or else use the method's default if available).
+        2. parameterized decoration:
+                @synpp.stage(bar='bar_pipeline')
+                def foo(bar): ...
+            Which will try to decide whether:
+             a. 'bar_pipeline' is an available stage, and pass the method its results, or
+             b. an argument in the global context config (or else use the method's default if available).
+             Note: the default value for 'bar' must be set in the actual function, not in the decorator.
+        3. fully parameterized decoration, which can be used for configuring the stage dependencies:
+                @synpp.stage(bar=synpp.stage('pipeline.bar', bar_argkey='bar_argval'))
+                def foo(bar): ...
+            Using synpp.stage() inside the decorator works in the same way as the decorator itself, simply pass the
+            target stage as first argument, and the config kwargs following it.
+        """
+        func_kwargs = {}
+        for name, param in self.func_params.items():
+            arg = self.stage_params.get(name)
+            if arg is None:
+                # case 1: simple decoration
+                config_kwargs = {'option': name}
+                if isinstance(context, ConfigurationContext) and param.default is not inspect._empty:
+                    config_kwargs.update({'default': param.default})
+                func_kwargs[name] = context.config(**config_kwargs)
+            elif resolve_stage(arg) is not None:
+                # case 2 or 3: stages only
+                func_kwargs[name] = context.stage(arg, config=getattr(arg, 'stage_params', {}))
+            else:
+                # case 2: config parameter
+                func_kwargs[name] = context.config(arg)
+
+        return func_kwargs
